#!/usr/bin/perl

# ReqTen - request-based mail forwarding script.
# $Id: main.pl,v 0.0.4 2006/02/30 07:20:00 diabah Exp $
#
# Copyright (c) 2006 diabah. All rights reserved.
# http://cubic9.com/Devel/Perl/reqten/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either versions 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA

use strict;
use warnings;
use POSIX qw(locale_h mktime strftime);
use Net::SMTP;
use MIME::Base64 qw(decode_base64);
use MIME::QuotedPrint qw(decode_qp);

setlocale(LC_TIME, 'C');

$0 =~ m#^(.*)/#;
my $script_dir = $1;

# load configuration file
open(IN, '< ' . $script_dir . '/main.conf')
  or &xdie('Failed to read configuration file.');
my $r_cfg = &parse_colon_sep_vals(<IN>);
close(IN);

&main;
exit;

sub main
{
    my $r_in_field = &parse_colon_sep_vals(<STDIN>);
    my $r_new_field;

    &filter_in_ml($r_in_field);

    $r_new_field = &make_new_ml($r_in_field);

    &send_new_ml($r_new_field);
}

sub parse_colon_sep_vals
{
    my $contents = join('', @_);

    my @header;
    my ($name, $value);
    my %field;

    # get body
    if ($contents =~ s/\n\n(.+)//s) {
        $field{'body'} = $1;
    }

    # get header
    $contents = &unfold_field($contents);
    @header = split(/\n/, $contents);
    foreach (@header) {
        ($name, $value) = split(/:\s*/, $_, 2);
        $field{lc($name)} = &fold_field($value);
    }

    return \%field;
}

sub filter_in_ml
{
    my $r_field = $_[0];

    my @ignorelist;

    if (
        (
            ($$r_field{'x-spam-status'})
         && ($$r_field{'x-spam-status'} =~ m/^Yes/i)
        ) || (
            ($$r_field{'x-proscan-status'})
         && ($$r_field{'x-proscan-status'} =~ m/^spam/i)
        )
       )
    {
        &xdie('spam.');
    }
    if (   ($$r_field{'x-virus-status'})
        && ($$r_field{'x-virus-status'} =~ m/^(Yes|Infected)/i))
    {
        &xdie('Virus.');
    }

    unless ($$r_field{'from'}) {
        &xdie('"From" is null.');
    }
    open(IN, '< ' . $script_dir . '/ignore.conf')
      or &xdie('Failed to read ignore list file.');
    @ignorelist = <IN>;
    close(IN);
    foreach (@ignorelist) {
        chomp($_);
        if (($_ ne '') && (index($$r_field{'from'}, $_) != -1)) {
            &xdie('"From" is in the ignore list.');
        }
    }
}

sub unfold_field
{
    my $tmp = $_[0];

    $tmp =~ s/\n\s+/\t \t \t/g;

    return $tmp;
}

sub fold_field
{
    my $tmp = $_[0];

    $tmp =~ s/\t \t \t/\n /g;

    return $tmp;
}

sub make_new_ml
{
    my $r_in_field = $_[0];

    my $r_ori_field;
    my %new_field;
    my $ori_filename;
    my @ori_lines;

    if (index($$r_in_field{'from'}, $$r_cfg{'receiver'}) == -1) {

        # in is normal ml
        $new_field{'message-id'} = $$r_in_field{'message-id'};
        $new_field{'message-id'} =~ s/</'<'.time.'reqten'/e;
        $new_field{'date'}       = &unix_to_rfc2822(time);
        $new_field{'from'}       = $$r_in_field{'from'};
        $new_field{'reply-to'}   = $$r_cfg{'sender'};
        $new_field{'to'}         = $$r_cfg{'receiver'};
        if ($$r_in_field{'message-id'}) {
            if ($$r_cfg{'receiver_add_r_field'}) {
                $new_field{'subject'} = 'reqten' . $$r_in_field{'subject'};
            } else {
                $new_field{'subject'} = 'reqten' . $$r_in_field{'message-id'};
                $new_field{'body'}    = &decode_subj($$r_in_field{'subject'});
            }
        } elsif ($$r_in_field{'date'}) {
            $new_field{'subject'} = 'reqten' . $$r_in_field{'date'};
            $new_field{'body'}    = &decode_subj($$r_in_field{'subject'});
        } else {
            &xdie('Both "Message-ID" and "Date" are null.');
        }
    } else {

        # stdin is request ml
        $ori_filename = &get_ori_filename($r_in_field);
        open(IN, '< ' . $ori_filename) or &xdie('Failed to read the mail.');
        @ori_lines = <IN>;
        close(IN);
        $r_ori_field = &parse_colon_sep_vals(@ori_lines);

        $new_field{'message-id'} = $$r_ori_field{'message-id'};
        $new_field{'message-id'} =~ s/</'<'.time.'reqten'/e;
        $new_field{'date'}       = $$r_ori_field{'date'};
        $new_field{'from'}       = $$r_ori_field{'from'};
        $new_field{'reply-to'}   = $$r_ori_field{'reply-to'};
        $new_field{'to'}         = $$r_cfg{'receiver'};
        $new_field{'subject'}    = $$r_ori_field{'subject'};

        #$new_field{'references'}  = $$r_in_field{'message-id'};
        #$new_field{'in-reply-to'} = $$r_in_field{'message-id'};
        $new_field{'content-type'} = $$r_ori_field{'content-type'};
        $new_field{'content-transfer-encoding'} =
          $$r_ori_field{'content-transfer-encoding'};
        $new_field{'body'} = $$r_ori_field{'body'};
    }

    return \%new_field;
}

sub unix_to_rfc2822
{
    my $timestamp = $_[0];

    return strftime('%a, %e %b %Y %H:%M:%S %z', localtime($timestamp));
}

# old encode_qp doesn't take 2nd arg :(
sub merge_line
{
    my $tmp = $_[0];

    $tmp =~ s/\n//g;

    return $tmp;
}

sub decode_subj
{
    my $subj = $_[0];

    $subj =~ s/\n\s*//g;
    $subj =~
      s#=\?ISO-2022-JP\?B\?([A-Za-z0-9\+/]+={0,2})\?=#decode_base64($1)#gei;
    $subj =~ s#=\?ISO-2022-JP\?Q\?([^ \t\?]+)\?=#decode_qp($1)#gei;

    return $subj;
}

sub get_ori_filename
{
    my $r_field = $_[0];

    my $in_subj;
    my $noti_msgid;
    my $key;
    my $ori_filename;

    $in_subj    = &decode_subj($$r_field{'subject'});
    $noti_msgid = $$r_field{'references'} || $$r_field{'in-reply-to'};

    unless ($in_subj =~ m/^Re:\s*reqten/i) {
        &xdie('It is not a request mail.');
    }

    if ($noti_msgid) {
        $key = $noti_msgid;
        $key =~ s/\d+reqten//;
        $ori_filename =
             &grep_ori_ml($$r_cfg{'maildir'} . '/new', 'message-id', $key)
          || &grep_ori_ml($$r_cfg{'maildir'} . '/cur', 'message-id', $key);
    } else {
        if ($in_subj =~ m/reqten(<.+)/) {
            $key          = $1;
            $ori_filename =
                 &grep_ori_ml($$r_cfg{'maildir'} . '/new', 'message-id', $key)
              || &grep_ori_ml($$r_cfg{'maildir'} . '/cur', 'message-id', $key);
        } elsif ($in_subj =~
            m/reqten((?:(Mon|Tue|Wed|Thu|Fri|Sat|Sun), )?(\d{1,2}) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) (\d{4}) (\d{2}):(\d{2})(?::(\d{2}))? (UT|GMT|[ECMP][SD]T|[ZAMNY]|[+-]\d{4}))/
          )
        {
            $key          = $1;
            $ori_filename =
                 &grep_ori_ml($$r_cfg{'maildir'} . '/new', 'date', $key)
              || &grep_ori_ml($$r_cfg{'maildir'} . '/cur', 'date', $key);
        } else {
            &xdie('Invalid request mail.');
        }
    }

    unless ($ori_filename) {
        &xdie('Failed to find the mail (grep key is ' . $key . ').');
    }

    return $ori_filename;
}

sub grep_ori_ml
{
    my ($dir, $field_name, $key) = @_;

    my @filelist;
    my $file;
    my @contents;

    # timestamp sort
    opendir(DIR, $dir);
    @filelist =
      map  { $_->[0] }
      sort { $a->[1] <=> $b->[1] }
      map  { [$_, -M $dir . '/' . $_] }
      grep(!/^\.\.?$/, readdir(DIR));
    closedir(DIR);

    foreach $file (@filelist) {
        open(IN, '< ' . $dir . '/' . $file)
          or &xdie('Failed to read a file in ' . $dir . '.');
        @contents = <IN>;
        close(IN);
        foreach (@contents) {
            if ($_ eq "\n") {
                # exit before rearch mail body
                last;
            } elsif (index(lc($_), $field_name . ':') == 0) {
                $_ =~ s/^$field_name:\s*//i;
                if (index($_, $key) == 0) {
                    return $dir . '/' . $file;
                }
            }
        }
    }

    return;
}

sub send_new_ml
{
    my $r_field = $_[0];

    my @field_seq = (
                     'Message-ID',
		     'Date',
                     'From',
		     'Reply-To',
                     'To',
		     'Subject',
                     'References',
		     'In-Reply-To',
                     'MIME-Version',
		     'Content-Type',
                     'Content-Transfer-Encoding'
                    );
    my $smtp = Net::SMTP->new($$r_cfg{'server'}, Timeout => 10, Debug => 1);

    $smtp->mail($$r_cfg{'sender'}) or &xdie('Failed to add the sender.');
    $smtp->to($$r_field{'to'})     or &xdie('Failed to add the recipient.');

    $smtp->data() or &xdie('Failed to add the body.');
    foreach (@field_seq) {
        if ($$r_field{lc($_)}) {
            $smtp->datasend($_ . ': ' . $$r_field{lc($_)} . "\n");
        }
    }
    $smtp->datasend("\n");
    $smtp->datasend($$r_field{'body'}) if ($$r_field{'body'});
    $smtp->dataend();

    $smtp->quit;
}

sub xdie
{
    my $msg = $_[0];

    print &unix_to_rfc2822(time) . ': ' . $msg . "\n";

    exit;
}
